#include <math.h>
#include <string>
#include <iostream>
#include <sstream>
#include <fstream>
#include <iomanip>
#include <numeric>
#include <exception>
#include <map>

#include "astrotime.h"
#include "coordinates.h"
#include "indices.h"
#include "parameters_set.h"
#include "processing.h"

using namespace boost::numeric::ublas;
// Preproccessing
// awk '{if($1=="O" || $1 == "P" || ($1 == "M" && $4=="Normal" )) print $0}' *.mass > *.e.mass
// cat *.e.mass *.stm > *.stmm
// sort -s -k 2,3 *.stmm > *.stm2 //The file for proccessing

const unsigned int data_begin = 22;
const unsigned int time_begin = 2;

unsigned int basetime_number = 0;

double mean_mz = 0.0;
double mjd_first = 0.0;
double mjd_last =  0.0;
Date dc_normal_mode;
Time ut_normal_mode;

void P_line( const std::string& s ) {
    if( basetime_number > 0 ){
        basetime_number = processing::run( dc_normal_mode, ut_normal_mode, mean_mz/basetime_number, 0.5*(mjd_first + mjd_last) );
    }
    parameters::P_line_parse( s.substr( 0, data_begin ), s.substr( data_begin ) );
}

void M_line( const std::string& s ) {
    Date dc;
    Time ut;
    std::string mode;
    std::istringstream mode_inp( s.substr( time_begin ) );
    mode_inp >> dc >> ut >> mode;
//    std::cerr << "M " << (mjd_last - mjd_first)*24*3600 << std::endl;
    if( ( mjd_last - mjd_first )*24*3600 + 1 > parameters::accumtime() && basetime_number > 0 ) {
        basetime_number = processing::run( dc_normal_mode, ut_normal_mode, mean_mz/basetime_number, 0.5*(mjd_first + mjd_last) );
    }
    if( mode == "Normal" && basetime_number == 0 ) {
        dc_normal_mode = dc;
        ut_normal_mode = ut;
    }
    if( mode == "Background" && basetime_number == 0 ) {
        std::cout << "M " << dc << ' ' << ut << " Background " << std::endl;
    }
}

void m_line( const std::string& s ) {
    Date dc;
    Time ut;
    std::istringstream stm_inp( s.substr( time_begin ) );
    stm_inp >> dc >> ut;
    mjd_last = (int)dc + ut/24.0;
    if( basetime_number == 0 ) mjd_first = mjd_last;
    if( basetime_number == 0 ) mean_mz = 0.0;
    mean_mz += processing::object_Mz( dc, ut );
    indices::calculation( s.substr( data_begin ) );
    basetime_number++;
}

void O_line( const std::string& s ) {
    if( basetime_number > 0 ){
        basetime_number = processing::run( dc_normal_mode, ut_normal_mode, mean_mz/basetime_number, 0.5*(mjd_first + mjd_last) );
    }
    std::istringstream tst( s.substr( data_begin ) );
    std::vector<std::string> items;
    tst >> items;
//    if( items.size() < 11 ) throw std::ios_base::failure( "Format error " );
    std::cout << "O " << s.substr( time_begin, data_begin-time_begin-1 ) << std::setw(5) << items.at(0);
    for( unsigned int i = 1; i < items.size() ; i++ ) std::cout << ' ' << items.at(i);
    std::cout << std::endl ;
    processing::object( items );
}

void I_line( const std::string& s ) {
//    std::string::size_type i = 0;
//    std::cerr << "I line" << std::endl;
}

void comment( const std::string& s ) {
    if( s.at(1) == '#' ) std::cout << s << std::endl;
}

namespace {
    typedef void(*CALLF_T)( const std::string&);
    typedef std::pair< char, CALLF_T> cfpair;
    cfpair file_line[] =  {
        cfpair( 'P', &P_line ),
        cfpair( 'M', &M_line ),
        cfpair( 'O', &O_line ),
        cfpair( 'I', &I_line ),
        cfpair( 'm', &m_line ),
        cfpair( '#', &comment ),
    };
}

static std::map< char, CALLF_T > func( file_line, file_line + sizeof(file_line)/sizeof(file_line[0]));


std::string headstamp( const std::string& pref ) {
    AstroDateTime utc;
    utc.current(CONV);
    std::ostringstream t( pref + ' ' + utc.strDate() + ' ' + utc.strTime(0) + ' ' );
    return t.str();
}

int main( int argc, char **argv ) {
    std::string mass_line;
    int line_number = 0;
    try {
        if( parameters::comand_line_parse( argc, argv  ) < 0 ) return -1;
        std::cout << headstamp( "#" ) << "Atmos-nnls Vers." << "2.96" << " Compiled " << __TIME__ << " " << __DATE__<< std::endl;
        parameters::init_keys();
        while( getline( std::cin, mass_line ) ) {
            try {
                line_number++;
                if( func.find( mass_line.at(0)) == func.end() ) continue;
                (*func[mass_line.at(0)])(mass_line);
            }
            catch (const std::ios_base::failure& e ) {
                std::cerr <<  mass_line.at(0) << "-line: " << e.what() << " at " << line_number << ". Skipped" << std::endl;
            }
            catch (const std::exception& e  ) {
                throw std::domain_error( std::string("Fatal error: ")+e.what() );
            }
        }
        if( basetime_number > 0 ){
            basetime_number = processing::run( dc_normal_mode, ut_normal_mode, mean_mz/basetime_number, 0.5*(mjd_first + mjd_last) );
        }
    }
    catch (const std::exception& e ) {
        std::cerr << e.what() << ". Line " << line_number << std::endl;
    }
    return 0;
}

